package org.bitbucket.bradleysmithllc.etlunit.maven;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.project.MavenProject;
import org.apache.maven.plugin.MojoExecutionException;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.feature.AbstractFeature;
import org.bitbucket.bradleysmithllc.etlunit.feature.ServiceLocatorFeatureLocator;
import org.bitbucket.bradleysmithllc.etlunit.feature.debug.ConsoleFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.feature.debug.RunAllFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.JSonBuilderProxy;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.List;

/**
 * Goal to run bitbucket tests
 *
 * @goal test
 * @phase test
 * @requiresDependencyResolution test
 */
public class                                                                                                                                                                          ETLUnitMojo
		extends AbstractMojo {
	/**
	 * @parameter expression="${project}" default-value="${project}"
	 */
	protected MavenProject mavenProject;

	/**
	 * @parameter expression="${project.version}" default-value="${project.version}"
	 */
	protected String projectVersion;

	/**
	 * @parameter expression="${project.name}" default-value="${project.name}"
	 */
	protected String projectName;

	/**
	 * @parameter expression="${outputDirectory}" default-value="${project.build.directory}/bitbucket-bin"
	 */
	protected File etlunitBinDirectory;

	/**
	 * @parameter expression="${sourceDirectory}" default-value="${basedir}/src"
	 */
	protected File srcDirectory;

	/**
	 * @parameter expression="${targetDirectory}" default-value="${basedir}/target/"
	 */
	protected File targetDirectory;

	/**
	 * @parameter expression="${etlunit.override.configuration}"
	 */
	protected String etlunitOverrideConfiguration;

	/**
	 * @parameter expression="${maven.test.skip}"
	 */
	protected String skip;

	/**
	 * @parameter expression="${rootDirectory}" default-value="${basedir}"
	 */
	protected File baseDirectory;

	public void execute()
			throws MojoExecutionException {
		if (skip != null) {
			// skip testing
			getLog().info("Test skipped");
			return;
		}

		if (etlunitOverrideConfiguration != null)
		{
			System.setProperty("etlunit.override.configuration", etlunitOverrideConfiguration);
		}
		else
		{
			System.clearProperty("etlunit.override.configuration");
		}

		File src = getEtlunitSourceDirectoryRoot(baseDirectory);

		if (!src.exists()) {
			src.mkdirs();
		}

		try {
			List classpathFiles = mavenProject.getTestClasspathElements();

			URL[] urls = new URL[classpathFiles.size()];

			getLog().debug("" + classpathFiles.size());

			for (int i = 0; i < classpathFiles.size(); ++i) {
				getLog().debug((String) classpathFiles.get(i));
				urls[i] = new File((String) classpathFiles.get(i)).toURL();
			}

			URLClassLoader ucl = new URLClassLoader(urls, Thread.currentThread().getContextClassLoader());

			ServiceLocatorFeatureLocator loc = new ServiceLocatorFeatureLocator(ucl);

			Configuration con = loadConfiguration(baseDirectory, projectName, projectVersion, ucl);

			final ETLTestVM vm = new ETLTestVM(loc, con);

			// 'activate' the console output
			ConsoleFeatureModule console = new ConsoleFeatureModule();
			vm.addFeature(console);

			// install a dummy feature so a director can be installed which will accept all tests
			vm.addFeature(new RunAllFeatureModule());

			// install a feature to create surefire reports
			vm.addFeature(new AbstractFeature() {
				private final StatusReporter statusReporter = new SurefireStatusReporter(vm.getRuntimeSupport().getReportDirectory("surefire"));

				public String getFeatureName() {
					return "surefire-reporter";
				}

				@Override
				public long getPriorityLevel() {
					return 1L;
				}

				@Override
				public StatusReporter getStatusReporter() {
					return statusReporter;
				}
			});

			vm.installFeatures();

			//TEST!
			TestResults results = vm.runTests();

			TestResultMetrics metrics = results.getMetrics();

			if (metrics.getNumberOfErrors() > 0 || metrics.getNumberOfAssertionFailures() > 0) {
				throw new MojoExecutionException("Build failed with test errors");
			}

			if (metrics.getNumberOfTestsPassed() == 0) {
				throw new MojoExecutionException("Build failed with no tests to run");
			}
		} catch (Exception e) {
			throw new MojoExecutionException("", e);
		}
	}

	/**
	 * These methods from here on are helpers for other modules which are operating in the maven project model
	 * and want to locate standard artifacts without having to spread that information all around the other modules.
	 *
	 * @return
	 */
	public static File getSourceDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("src").mkdirs().file();
	}

	public static Configuration loadConfiguration(
			File basedir,
			String projectName,
			String projectVersion,
			ClassLoader urlLoader
	) throws IOException {
		File src = new File(basedir, "src");
		File test = new File(src, "test");
		File testResources = new File(test, "resources");

		File cfg = new File(testResources, "config");

		final Configuration con;
		try {
			ETLTestValueObject base =
					ETLTestParser.loadObject(
							new JSonBuilderProxy()
									.object()
									.key("project-root-directory")
									.value(basedir.getAbsolutePath())
									.key("vendor-binary-directory")
									.value(getBinDirectoryRoot(basedir).getAbsolutePath())
									.key("test-sources-directory")
									.value(getEtlunitSourceDirectoryRoot(basedir))
									.key("generated-source-directory")
									.value(getGeneratedSourceDirectoryRoot(basedir))
									.key("temp-directory")
									.value(getTempDirectoryRoot(basedir))
									.key("reports-directory")
									.value(getReportsDirectoryRoot(basedir))
									.key("resource-directory")
									.value(getResourceDirectory(basedir))
									.key("reference-directory")
									.value(getReferenceDirectory(basedir))
									.key("configuration-directory")
									.value(getConfigurationRoot(basedir))
									.key("project-version")
									.value(projectVersion)
									.key("project-name")
									.value(projectName)
									.key("project-user")
									.value(System.getProperty("user.name"))
									.endObject()
									.toString()
			);

			// check for an override in the system properties.  Default is none
			String override = System.getProperty("etlunit.override.configuration");

			// merge together, preferring the override
			con = Configuration.loadFromEnvironment(cfg, override, urlLoader, base);


			// save off the resulting config file for reference
			File configTarget = new FileBuilder(getGeneratedSourceDirectoryRoot(basedir)).subdir("config").mkdirs().name("effective_configuration.json").file();

			IOUtils.writeBufferToFile(configTarget, new StringBuffer(con.getRoot().getJsonNode().toString()));

			return con;
		} catch (ParseException e) {
			throw new IOException(e);
		}
	}

	protected static File getResourceDirectory(File basedir) {
		return getMainDirectory(basedir);
	}

	public static File getEtlunitSourceDirectoryRoot(File basedir) {
		return new FileBuilder(getTestDirectory(basedir)).subdir("etlunit").mkdirs().file();
	}

	public static File getBuildDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("target").mkdirs().file();
	}

	public static File getGeneratedSourceDirectoryRoot(File project_root) {
		return new FileBuilder(getBuildDirectoryRoot(project_root)).subdir("generated-sources").mkdirs().file();
	}

	public static File getBinDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("target").subdir("etlunit-bin").mkdirs().file();
	}

	public static File getReportsDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("target").mkdirs().file();
	}

	public static File getLibDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("target").subdir("etlunit-lib").mkdirs().file();
	}

	public static File getTempDirectoryRoot(File project_root) {
		return new FileBuilder(project_root).subdir("target").subdir("etlunit-temp").mkdirs().file();
	}

	public static File getEtlunitConfiguration(File project_root) {
		return new FileBuilder(getConfigurationRoot(project_root)).mkdirs().name("etlunit.json").file();
	}

	public static File getConfigurationRoot(File project_root) {
		return new FileBuilder(getTestDirectory(project_root)).subdir("resources").subdir("config").mkdirs().file();
	}

	public static File getFeatureSource(File project_root, String feature) {
		return new FileBuilder(getSourceDirectoryRoot(project_root)).subdir("main").subdir(feature).mkdirs().file();
	}

	public static File getMainDirectory(File project_root) {
		return new FileBuilder(getSourceDirectoryRoot(project_root)).subdir("main").mkdirs().file();
	}

	public static File getReferenceDirectory(File project_root) {
		return new FileBuilder(getMainDirectory(project_root)).subdir("reference").mkdirs().file();
	}

	public static File getTestDirectory(File project_root) {
		return new FileBuilder(getSourceDirectoryRoot(project_root)).subdir("test").mkdirs().file();
	}

	public static File getFeatureConfiguration(File project_root, String feature) {
		return new FileBuilder(getConfigurationRoot(project_root)).subdir(feature).mkdirs().file();
	}
}